import re

import validators
from base import ARMObject, ARMProperty, SubResource
from network import BackendAddressPoolRef, ProbeRef


class Plan(ARMProperty):
    props = {
        'name': (str, False),
        'publisher': (str, False),
        'product': (str, False),
        'promotionCode': (str, False)
    }


class HardwareProfile(ARMProperty):
    props = {
        'vmSize': (str, False)
    }
    # todo add validations


class ImageReference(ARMProperty):
    def __init__(self, title=None, version='latest', **kwargs):
        kwargs['version'] = version
        ARMProperty.__init__(self, title, **kwargs)

    # todo add validations
    props = {
        'publisher': (str, False),
        'offer': (str, False),
        'sku': (str, False),
        'version': (str, False)
    }


class VirtualHardDisk(ARMProperty):
    props = {
        'uri': (str, False)
    }


class ManagedDiskParameters(ARMProperty):
    props = {
        'id': (str, False),
        'storageAccountType': (str, False)  # todo add validation - Standard_LRS or Premium_LRS
    }

    def validate(self):
        if 'storageAccountType' in self.properties:
            if self.properties['storageAccountType'] not in ['Standard_LRS', 'Premium_LRS']:
                raise ValueError('storageAccountType - Possible values are: Standard_LRS or Premium_LRS')


# https://docs.microsoft.com/en-us/azure/templates/microsoft.compute/virtualmachines#OSDisk
class OsDisk(ARMProperty):
    props = {
        'osTpe': (str, False),  # todo add validations: Linux/Windows
        # 'encryptionSettings': (object, False),  # todo add support for encryptionSettings
        'name': (str, False),
        'vhd': (VirtualHardDisk, False),
        'image': (VirtualHardDisk, False),
        # 'caching': (str, False),  # todo add support
        'createOption': (str, True),  # todo add validations
        'diskSizeGB': (int, False),  # todo add validations on max size
        'managedDisk': (ManagedDiskParameters, False)
    }


class StorageProfile(ARMProperty):
    props = {
        'imageReference': (ImageReference, False),
        'osDisk': (OsDisk, False)
        # 'dataDisks':  # todo add support
    }


class WindowsConfiguration(ARMProperty):
    props = {
        'provisionVMAgent': (bool, False),  # default is True
        'enableAutomaticUpdates': (bool, False),
        'timeZone': (str, False),
        # 'additionalUnattendContent': (str, False),  # todo add support
        # 'winRM': (str, False)  # todo add support
    }


class SshPublicKey(ARMProperty):
    props = {
        'path': (str, False),
        'keyData': (str, False)
    }


class SshConfiguration(ARMProperty):
    props = {
        'publicKeys': ((list, SshPublicKey), False)
    }


class LinuxConfiguration(ARMProperty):
    props = {
        'disablePasswordAuthentication': (bool, False),
        'ssh': (SshConfiguration, False)
    }


class OSProfile(ARMProperty):
    # todo add valiation: https://docs.microsoft.com/en-us/azure/templates/microsoft.compute/virtualmachines#OSProfile
    props = {
        'computerName': (str, False),
        'adminUsername': (str, False),
        'adminPassword': (str, False),
        'customData': (str, False),
        'windowsConfiguration': (WindowsConfiguration, False),
        'linuxConfiguration': (LinuxConfiguration, False)
        # 'secrets': (, False)  # todo add support
    }


class NetworkInterfaceReferenceProperties(ARMProperty):
    props = {
        'primary': (bool, False)
    }


class NetworkInterfaceReference(ARMProperty):
    props = {
        'id': (str, False),
        'properties': (NetworkInterfaceReferenceProperties, False)
    }


class NetworkProfile(ARMProperty):
    props = {
        'networkInterfaces': ([NetworkInterfaceReference], False)
    }


class BootDiagnostics(ARMProperty):
    props = {
        'enabled': (bool, False),
        'storageUri': (str, False)
    }


class DiagnosticsProfile(ARMProperty):
    props = {
        'bootDiagnostics': (BootDiagnostics, False)
    }


valid_names_azure_extensions = re.compile(r'^[a-zA-Z0-9_\-/]+$')


class VirtualMachineExtension(ARMObject):
    # name of extension should be "[concat(parameters('vmName'),'/parameters('extensionName')]"
    # and extension should be dependent on a VM
    resource_type = 'Microsoft.Compute/virtualMachines/extensions'
    apiVersion = "2017-12-01"
    location = True

    props = {
        'forceUpdateTag': (bool, False),
        'publisher': (str, False),
        'type': (str, False),
        'typeHandlerVersion': (str, False),
        'autoUpgradeMinorVersion': (bool, False),
        'settings': (dict, False),
        'protectedSettings': (dict, False)
    }

    root_props = {
        'tags': (dict, False)
    }

    def validate_title(self):
        if not valid_names_azure_extensions.match(self.title):
            raise ValueError('Name "%s" is not valid' % self.title)

    @staticmethod
    def create_linux_custom_script(vm, command_to_execute, script=None, file_uris=None, storage_account_name=None,
                                   storage_account_key=None, protected_settings=True, tags=None):

        vm_extention = VirtualMachineExtension(vm.title + '/CustomScriptForLinux',
                                               publisher='Microsoft.Azure.Extensions',
                                               type='CustomScript',
                                               autoUpgradeMinorVersion=True,
                                               typeHandlerVersion='2.0',
                                               dependsOn=vm)

        VirtualMachineExtension._set_common_extension_values(vm_extention=vm_extention,
                                                             vm=vm,
                                                             command_to_execute=command_to_execute,
                                                             script=script,
                                                             file_uris=file_uris,
                                                             storage_account_name=storage_account_name,
                                                             storage_account_key=storage_account_key,
                                                             protected_settings=protected_settings,
                                                             tags=tags)

        return vm_extention

    @staticmethod
    def create_windows_custom_script(vm, command_to_execute, script=None, file_uris=None, storage_account_name=None,
                                     storage_account_key=None, protected_settings=True, tags=None):

        vm_extention = VirtualMachineExtension(vm.title + '/CustomScriptForWindows',
                                               publisher='Microsoft.Compute',
                                               type='CustomScriptExtension',
                                               autoUpgradeMinorVersion=True,
                                               typeHandlerVersion='1.9',
                                               dependsOn=vm)

        VirtualMachineExtension._set_common_extension_values(vm_extention=vm_extention,
                                                             vm=vm,
                                                             command_to_execute=command_to_execute,
                                                             script=script,
                                                             file_uris=file_uris,
                                                             storage_account_name=storage_account_name,
                                                             storage_account_key=storage_account_key,
                                                             protected_settings=protected_settings,
                                                             tags=tags)

        return vm_extention

    @staticmethod
    def _set_common_extension_values(vm_extention, vm, command_to_execute, script=None, file_uris=None,
                                     storage_account_name=None, storage_account_key=None, protected_settings=True,
                                     tags=None):

        if script and file_uris:
            raise ValueError("Must set either script or an array of file uris")
        if not isinstance(vm, VirtualMachine):
            raise ValueError("Must provide a VirtualMachine object")

        settings = {'commandToExecute': command_to_execute}

        if script:
            settings['script'] = script

        if file_uris:
            if isinstance(file_uris, str):
                file_uris = [file_uris]
            settings['fileUris'] = file_uris

        if storage_account_key:
            settings['storageAccountKey'] = storage_account_key

        if storage_account_name:
            settings['storageAccountName'] = storage_account_name

        if tags:
            vm_extention.properties['tags'] = tags

        if protected_settings:
            vm_extention.properties['protectedSettings'] = settings
        else:
            vm_extention.properties['settings'] = settings

        vm_extention.with_depends_on(vm)


class VirtualMachine(ARMObject):
    resource_type = 'Microsoft.Compute/virtualMachines'
    apiVersion = "2017-12-01"
    location = True

    props = {
        'hardwareProfile': (HardwareProfile, False),
        'storageProfile': (StorageProfile, False),
        'osProfile': (OSProfile, False),
        'networkProfile': (NetworkProfile, False),
        'diagnosticsProfile': (DiagnosticsProfile, False),
        # 'availabilitySet': (),  # todo add support
        # 'licenseType': ()  # todo add support
    }

    root_props = {
        'plan': (Plan, False),
        'tags': (dict, False)
    }


class RollingUpgradePolicy(ARMProperty):
    props = {
        'maxBatchInstancePercent': (int, False),
        'maxUnhealthyInstancePercent': (int, False),
        'maxUnhealthyUpgradedInstancePercent': (int, False),
        'pauseTimeBetweenBatches': (str, False)
    }


class UpgradePolicy(ARMProperty):
    props = {
        'mode': (str, False),  # Manual or Automatic
        'rollingUpgradePolicy': (RollingUpgradePolicy, False),
        'automaticOSUpgrade': (bool, False)
    }


# https://docs.microsoft.com/en-us/azure/templates/microsoft.compute/virtualmachinescalesets#VirtualMachineScaleSetOSProfile
class VirtualMachineScaleSetOSProfile(ARMProperty):
    props = {
        'computerNamePrefix': (str, True),
        'adminUsername': (str, True),
        'adminPassword': (str, True),
        'customData': (str, False),
        'windowsConfiguration': (WindowsConfiguration, False),
        'linuxConfiguration': (LinuxConfiguration, False),
        # 'secrets': (list, False),  - not implemented
    }

    def validate(self):
        if 'computerNamePrefix' in self.properties:
            val = self.properties['computerNamePrefix']
            if len(val) < 0 or len(val) > 15:
                raise ValueError("computerNamePrefix - Computer name prefixes must be 1 to 15 characters long.")


# https://docs.microsoft.com/en-us/azure/templates/microsoft.compute/virtualmachinescalesets#VirtualMachineScaleSetOSDisk
class VirtualMachineScaleSetOSDisk(ARMProperty):
    props = {
        'name': (str, False),
        'caching': (str, False),
        'createOption': (str, True),
        'osType': (str, False),
        'image': (VirtualHardDisk, False),
        'vhdContainers': (list, False),
        'managedDisk': (ManagedDiskParameters, False),
    }


class VirtualMachineScaleSetDataDisk(ARMProperty):
    props = {
        'name': (str, False),
        'lun': (int, False),
        'caching': (str, False),
        'createOption': (str, True),
        'diskSizeGB': (int, False),
        'managedDisk': (ManagedDiskParameters, False),
    }


class VirtualMachineScaleSetStorageProfile(ARMProperty):
    props = {
        'imageReference': (ImageReference, False),
        'osDisk': (VirtualMachineScaleSetOSDisk, False),
        'dataDisks': (list, False)  # VirtualMachineScaleSetDataDisk
    }


class VirtualMachineScaleSetPublicIPAddressConfigurationDnsSettings(ARMProperty):
    props = {
        'domainNameLabel': (str, True)
    }


class VirtualMachineScaleSetPublicIPAddressConfiguration(ARMObject):
    # title must be provided
    props = {
        'idleTimeoutInMinutes': (int, False),
        'dnsSettings': (VirtualMachineScaleSetPublicIPAddressConfigurationDnsSettings, False)
    }


class VirtualMachineScaleSetIPConfiguration(ARMObject):
    # title must be provided
    props = {
        'id': (str, False),
        'subnet': (SubResource, False),
        'primary': (bool, False),
        'publicIPAddressConfiguration': (VirtualMachineScaleSetPublicIPAddressConfiguration, False),
        'privateIPAddressVersion': (SubResource, False),  # Possible values are: 'IPv4' and 'IPv6'
        'applicationGatewayBackendAddressPools': ([SubResource], False),
        'loadBalancerBackendAddressPools': ([BackendAddressPoolRef], False),
        'loadBalancerInboundNatPools': ([SubResource], False)
    }

    def to_dict(self):
        self._move_prop_to_root('id')
        return ARMObject.to_dict(self)


class VirtualMachineScaleSetNetworkConfiguration(ARMObject):
    # title must be provided (will be converted to name internally)
    props = {
        'id': (str, False),
        'primary': (bool, False),
        'enableAcceleratedNetworking': (bool, False),
        'networkSecurityGroup': (SubResource, False),
        # 'dnsSettings': (object, False), todo - not implemented
        'ipConfigurations': ([VirtualMachineScaleSetIPConfiguration], False),
        'enableIPForwarding': (bool, False)
    }


class VirtualMachineScaleSetNetworkProfile(ARMProperty):
    props = {
        'healthProbe': (ProbeRef, False),  # this feature is in preview and requires special feature
        'networkInterfaceConfigurations': ([VirtualMachineScaleSetNetworkConfiguration], False)
    }


class VirtualMachineScaleSetExtensionProperties(ARMProperty):
    props = {
        'forceUpdateTag': (str, False),
        'publisher': (str, False),
        'type': (str, False),  # example "CustomScriptExtension"
        'typeHandlerVersion': (str, False),
        'autoUpgradeMinorVersion': (bool, False),
        'settings': (dict, False),
        'protectedSettings': (dict, False)
    }


class VirtualMachineScaleSetExtension(ARMProperty):
    props = {
        'name': (str, False),
        'properties': (VirtualMachineScaleSetExtensionProperties, True)
    }


class VirtualMachineScaleSetExtensionProfile(ARMProperty):
    props = {
        'extensions': ([VirtualMachineScaleSetExtension], True)
    }


class VirtualMachineScaleSetVMProfile(ARMProperty):
    props = {
        'osProfile': (VirtualMachineScaleSetOSProfile, False),
        'storageProfile': (VirtualMachineScaleSetStorageProfile, False),
        'networkProfile': (VirtualMachineScaleSetNetworkProfile, False),
        # 'diagnosticsProfile': (DiagnosticsProfile, False), todo - not implemented
        'extensionProfile': (VirtualMachineScaleSetExtensionProfile, False),
        'licenseType': (str, False)
    }


class VirtualMachineScaleSetSku(ARMProperty):
    props = {
        'name': (str, True),
        'tier': (str, True),  # Standard or Basic
        'capacity': (int, True),
    }


# https://docs.microsoft.com/en-us/azure/templates/microsoft.compute/virtualmachinescalesets#microsoftcomputevirtualmachinescalesets-object
class VirtualMachineScaleSets(ARMObject):
    resource_type = 'Microsoft.Compute/virtualMachineScaleSets'
    apiVersion = "2017-12-01"
    location = True

    props = {
        # 'identity': (object, False),   # root prop - not implemented
        'zones': ([str], False),
        'upgradePolicy': (UpgradePolicy, True),
        'virtualMachineProfile': (VirtualMachineScaleSetVMProfile, False),
        'overprovision': (bool, False),
        'singlePlacementGroup': (bool, False),
    }

    root_props = {
        'sku': (VirtualMachineScaleSetSku, False),
        'plan': (Plan, False),
        'tags': (dict, False)
    }

    def to_dict(self):
        self._move_prop_to_root('zones')
        return ARMObject.to_dict(self)



#!/usr/bin/python

from ionosphere import Base64, Join


def from_file(filepath, delimiter='', blanklines=False):
    """
    Imports userdata from a file.

    :type filepath: string
    :param filepath
    The absolute path to the file.

    :type delimiter: string
    :param: delimiter
    Delimiter to use with the ionosphere.Join().

    :type blanklines: boolean
    :param blanklines
    If blank lines shoud be ignored

    rtype: ionosphere.Base64
    :return The base64 representation of the file.
    """

    data = []

    try:
        with open(filepath, 'r') as f:
            for line in f:
                if blanklines and line.strip('\n\r ') == '':
                    continue

                data.append(line)
    except IOError:
        raise IOError('Error opening or reading file: {}'.format(filepath))

    return Base64(Join(delimiter, data))



from compute import *

from base import ARMTemplate
from network import *

template = ARMTemplate()

vnet = VirtualNetwork("myvnet",
                      addressSpace=AddressSpace(addressPrefixes=['10.0.0.0/24']),
                      subnets=[Subnet('subnet1', addressPrefix='10.0.0.0/24')])

#region load balancer

# needs to be Static so it will be assigned a value immediately
lb_pub_ip = PublicIPAddress('lb_pub_ip', publicIPAllocationMethod='Static')

lb_ip_conf = FrontendIPConfiguration('lb_ip_conf',
                                     privateIPAllocationMethod='Dynamic',
                                     # subnet=SubResource(id=vnet.SubnetRef('subnet1')),
                                     publicIPAddress=SubResource(id=lb_pub_ip.ref()))

lb_rule = LoadBalancingRule('lbrule',
                            protocol='Tcp',
                            loadDistribution='SourceIP',
                            frontendPort=22,
                            backendPort=22,
                            frontendIPConfiguration=FrontendIPConfigurationRef(load_balancer='my_lb',
                                                                               ip_conf=lb_ip_conf),
                            probe=ProbeRef(load_balancer='my_lb', probe='testProbe'),
                            backendAddressPool=BackendAddressPoolRef(load_balancer='my_lb', backend_pool='lb_backend'))

probe = Probe('testProbe', protocol='Tcp', port=22)

lb = LoadBalancer('my_lb',
                  sku=LoadBalancerSku(name='Basic'),
                  frontendIPConfigurations=[lb_ip_conf],
                  backendAddressPools=[BackendAddressPool(name='lb_backend')],
                  probes=[probe],
                  loadBalancingRules=[lb_rule])
lb.with_depends_on([lb_pub_ip, vnet])

#endregion


#region scaleset

linux_conf = LinuxConfiguration(disablePasswordAuthentication=False)

ip_conf = VirtualMachineScaleSetIPConfiguration('my_ip_config',
                                                subnet=SubResource(id=vnet.SubnetRef('subnet1')),
                                                loadBalancerBackendAddressPools=[BackendAddressPoolRef(load_balancer=lb,
                                                                                                       backend_pool='lb_backend')])
nic_conf = VirtualMachineScaleSetNetworkConfiguration('my_nic',
                                                      primary=True,
                                                      ipConfigurations=[ip_conf])

ss_vm_prof = VirtualMachineScaleSetVMProfile(
    osProfile=VirtualMachineScaleSetOSProfile(computerNamePrefix='testVm',
                                              adminUsername='adminuser',
                                              adminPassword='Welcome1234567+',
                                              linuxConfiguration=linux_conf),
    storageProfile=VirtualMachineScaleSetStorageProfile(imageReference=ImageReference(publisher='Canonical',
                                                                                      offer='UbuntuServer',
                                                                                      sku='16.04-LTS'),
                                                        osDisk=VirtualMachineScaleSetOSDisk(createOption='FromImage')),
    networkProfile=VirtualMachineScaleSetNetworkProfile(networkInterfaceConfigurations=[nic_conf]))

vm_scale_set = VirtualMachineScaleSets('my_scale_set',
                                       # plan=Plan(),
                                       sku=VirtualMachineScaleSetSku(name='Standard_A1', tier='Standard', capacity=2),
                                       overprovision=True,
                                       upgradePolicy=UpgradePolicy(mode='Manual'),
                                       virtualMachineProfile=ss_vm_prof).with_depends_on([vnet, lb])

#endregion


template.add_resource([lb_pub_ip, lb, vm_scale_set, vnet])

print(template.to_json())



from base import ARMObject, ARMProperty


class StorageAccountSku(ARMProperty):
    props = {
        'name': (str, True)  # todo - add validation: Gets or sets the sku name.
        # todo - Required for account creation; optional for update
        # todo - values: Standard_LRS, Standard_GRS, Standard_RAGRS, Standard_ZRS, Premium_LRS
    }


class VirtualNetworkRule(ARMProperty):
    props = {
        'id': (str, True)
    }


class IPRule(ARMProperty):
    props = {
        'value': (str, True)  # Specifies the IP or IP range in CIDR format. Only IPV4 address is allowed.
    }


class NetworkRuleSet(ARMProperty):
    props = {
        # 'bypass': (str, False)
        'virtualNetworkRules': ((list, VirtualNetworkRule), False),
        'ipRules': ((list, IPRule), False),
        'defaultAction': (str, True)  # Allow/Deny
    }


# todo add title validaton: 3-24 chars lengths, numbers and lower case letters
class StorageAccount(ARMObject):
    resource_type = 'Microsoft.Storage/storageAccounts'
    apiVersion = "2017-10-01"
    location = True

    props = {
        # 'customDomain':
        # 'encryption':
        'networkAcls': (NetworkRuleSet, False),
        'supportsHttpsTrafficOnly': (bool, False),
        'accessTier': (str, False)
    }

    root_props = {
        'sku': (StorageAccountSku, True),
        'kind': (str, True),  # Required. Indicates the type of storage account. - Storage, StorageV2, BlobStorage
        'tags': (dict, False)
    }



import json
import re

import cfn_flip as cfn_flip

from ionosphere import encode_to_dict, AWSObject, AWSProperty, Parameter, validators, AWSHelperFn

# Template Limits
MAX_VARIABLES = 256
MAX_OUTPUTS = 64
MAX_PARAMETERS = 256
MAX_RESOURCES = 800
PARAMETER_TITLE_MAX = 255

valid_names_azure = re.compile(r'^[a-zA-Z0-9_\-]+$')


class ARMTemplate(object):
    props = {
        '$schema': (str, False),
        'contentVersion': (str, False),
        'variables': (str, False),
        'parameters': (dict, False),
        'resources': (dict, True),
        'outputs': (dict, False),
    }

    def __init__(self, contentVersion="1.0.0.0"):
        self.contentVersion = contentVersion
        self.variables = []
        self.parameters = {}
        self.resources = []
        self.outputs = {}

    def handle_duplicate_key(self, key):
        raise ValueError('duplicate key "%s" detected' % key)

    def _update(self, d, values):
        if isinstance(values, list):
            for v in values:
                if v.title in d:
                    self.handle_duplicate_key(v.title)
                d[v.title] = v
        else:
            if values.title in d:
                self.handle_duplicate_key(values.title)
            d[values.title] = values
        return values

    def _update_list(self, lst, values):
        if isinstance(values, list):
            for v in values:
                item = next(iter(filter(lambda x: hasattr(x, 'title') and hasattr(v, 'title') and x.title == v.title,
                                        lst)), None)
                if item:
                    self.handle_duplicate_key(v.title)
                else:
                    lst.append(v)
        else:
            item = next(
                iter(filter(lambda x: hasattr(x, 'title') and hasattr(values, 'title') and x.title == values.title, lst)),
                None)
            if item:
                self.handle_duplicate_key(values.title)
            else:
                lst.append(values)
        return values

    def add_output(self, output):
        if len(self.outputs) >= MAX_OUTPUTS:
            raise ValueError('Maximum outputs %d reached' % MAX_OUTPUTS)
        return self._update(self.outputs, output)

    def add_parameter(self, parameter):
        if len(self.parameters) >= MAX_PARAMETERS:
            raise ValueError('Maximum parameters %d reached' % MAX_PARAMETERS)
        return self._update(self.parameters, parameter)

    def add_variable(self, variable):
        if len(self.variables) >= MAX_VARIABLES:
            raise ValueError('Maximum variables %d reached' % MAX_VARIABLES)
        return self._update_list(self.variables, variable)

    def add_resource(self, resource):
        if len(self.resources) >= MAX_RESOURCES:
            raise ValueError('Maximum number of resources %d reached'
                             % MAX_RESOURCES)
        return self._update_list(self.resources, resource)

    def add_content_version(self, version=None):
        if version:
            self.contentVersion = version
        else:
            self.contentVersion = "1.0.0.0"

    def to_dict(self):
        t = {}

        if self.outputs:
            t['outputs'] = self.outputs
        if self.parameters:
            t['parameters'] = self.parameters
        if self.contentVersion:
            t['contentVersion'] = self.contentVersion

        t['resources'] = self.resources
        t['$schema'] = "http://schema.management.azure.com/schemas/2015-01-01/deploymentTemplate.json"

        return encode_to_dict(t)

    def to_json(self, indent=4, sort_keys=True, separators=(',', ': ')):
        return json.dumps(self.to_dict(), indent=indent,
                          sort_keys=sort_keys, separators=separators)

    def to_yaml(self, long_form=False):
        return cfn_flip.to_yaml(self.to_json(), long_form)


class ARMObject(AWSObject):
    dictname = 'properties'
    root_props = None

    def __init__(self, title, template=None, validation=True, **kwargs):
        AWSObject.__init__(self, title, template, validation, **kwargs)

        if hasattr(self, 'location') and self.location and 'location' not in self.resource:
            self.resource['location'] = "[resourceGroup().location]"

        if title:
            self.resource['name'] = self.title

        if hasattr(self, 'apiVersion'):
            self.resource['apiVersion'] = self.apiVersion

        # move root props from properties to resources dict to be in root of object when serialized
        # super(ARMObject, self)._validate_props()
        # for rootProp in list(filter(lambda x: issubclass(type(x[1]), ARMRootProperty), self.properties.items())):
        #     k, v = rootProp
        #     self.resource[k] = v
        #     del self.properties[k]
        #     del self.props[k]

    def to_dict(self):
        if 'name' in self.resource:
            self.resource['name'] = self.title
        if 'tags' in self.properties:
            self.resource['tags'] = self.properties['tags']
            del self.properties['tags']
        if 'dependsOn' in self.properties:
            self.resource['dependsOn'] = self.properties['dependsOn']
            del self.properties['dependsOn']
        return AWSObject.to_dict(self)

    def ref(self):
        return "[resourceId('{0}/','{1}')]".format(self.resource['type'], self.title)

    Ref = ref

    def get_att(self, value):
        raise NotImplementedError()

    # todo - add stuff relevant for azure like refs to properties shortcuts similar to what we have in in the AWSObject

    def validate_title(self):
        if not valid_names_azure.match(self.title):
            raise ValueError('Name "%s" is not valid' % self.title)

    # fluent api
    def with_depends_on(self, depends_on):
        if 'dependsOn' not in self.resource:
            self.resource['dependsOn'] = []
        self.resource['dependsOn'].extend(self._add_dependencies(depends_on))
        return self

    def _add_dependencies(self, value):
        if isinstance(value, list):
            result = []
            for d in value:
                result.append(self._add_dependency(d))
            return result
        else:
            return [self._add_dependency(value)]

    def _add_dependency(self, dependency):
        if isinstance(dependency, ARMObject):
            return dependency.Ref()
        elif isinstance(dependency, str):
            return dependency
        else:
            raise ValueError('Unsupported type')

    def _move_prop_to_root(self, key):
        if key in self.properties:
            self.resource[key] = self.properties[key]
            del self.properties[key]


class ARMProperty(AWSProperty):
    # Used to specify properties that should be on the root object in Azure
    root_props = None

    def validate_title(self):
        if not valid_names_azure.match(self.title):
            raise ValueError('Name "%s" is not valid' % self.title)


# class ARMRootProperty(ARMProperty):
#     pass


class ARMParameter(Parameter):
    STRING_PROPERTIES = ['maxLength', 'minLength']
    NUMBER_PROPERTIES = ['maxValue', 'minValue']
    SUPPORTED_TYPES = ['string', 'secureString', 'int', 'bool', 'object', 'secureObject', 'array']

    props = {
        'type': (str, True),
        'defaultValue': ((str, int, float), False),
        'allowedValues': (list, False),
        'maxLength': (validators.positive_integer, False),
        'minLength': (validators.positive_integer, False),
        'maxValue': (validators.integer, False),
        'minValue': (validators.integer, False),
        'description': (str, False)
    }

    root_props = None

    def to_dict(self):
        if 'description' in self.properties:
            description = self.properties['description']
            self.properties['metadata'] = {'description': description}
            del self.properties['description']

        return Parameter.to_dict(self)

    def ref(self):
        return "[parameters('{}')]".format(self.title)

    Ref = ref

    def validate(self):
        def check_type(t, v):
            try:
                t(v)
                return True
            except ValueError:
                return False

        # Validate allowed types
        if self.properties['type'] not in self.SUPPORTED_TYPES:
            raise ValueError("{} is not a supported parameter type. Supported types: {}".format(
                self.properties['type'], self.SUPPORTED_TYPES))

        # Validate the Default parameter value
        default = self.properties.get('defaultValue')
        if default:
            error_str = ("Parameter default type mismatch: expecting "
                         "type %s got %s with value %r")
            # Get the Type specified and see whether the default type
            # matches (in the case of a String Type) or can be coerced
            # into one of the number formats.
            param_type = self.properties.get('type')
            if (param_type == 'string' or param_type == 'secureString') and not isinstance(default, str):
                raise ValueError(error_str %
                                 ('string', type(default), default))
            elif param_type == 'int' and not isinstance(default, int):
                raise ValueError(error_str %
                                 (param_type, type(default), default))
            elif param_type == 'bool' and not isinstance(default, bool):
                raise ValueError(error_str %
                                 (param_type, type(default), default))
            elif param_type == 'array':
                if not isinstance(default, str):
                    raise ValueError(error_str %
                                     (param_type, type(default), default))
                allowed = [str, int]
                dlist = default.split(",")
                for d in dlist:
                    # Verify the split array are all in allowed
                    if not any([check_type(x, d) for x in allowed]):
                        raise ValueError(error_str %
                                         (param_type, type(d), dlist))

        if self.properties['type'] != 'string':
            for p in self.STRING_PROPERTIES:
                if p in self.properties:
                    raise ValueError("%s can only be used with parameters of "
                                     "the String type." % p)
        if self.properties['type'] != 'int':
            for p in self.NUMBER_PROPERTIES:
                if p in self.properties:
                    raise ValueError("%s can only be used with parameters of "
                                     "the Number type." % p)


class SubResource(ARMProperty):
    props = {
        'id': (str, True),
    }


class SubResourceRef(AWSHelperFn):
    def __init__(self, root_resource, sub_resource, root, child):
        sub_resource_ref = self._build_sub_resource_ref(root_resource, sub_resource)
        root_name = self._get_title(root)
        child_name = self._get_title(child)
        self.data = {'id': "[resourceId('{0}', '{1}', '{2}')]".format(sub_resource_ref, root_name, child_name)}

    def _build_sub_resource_ref(self, root_resource, sub_resource):
        root_resource_type = self._get_resource_type(root_resource)
        sub_resource_type = self._get_resource_type(sub_resource)
        return "{0}/{1}".format(root_resource_type, sub_resource_type)

    def _get_resource_type(self, r):
        return r if isinstance(r, str) else r.resource_type

    def _get_title(self, obj):
        if isinstance(obj, str):
            return obj
        return obj.title



#
# Regions
#

# todo add region consts


#
# Networking
#

QUAD_ZERO = '0.0.0.0/0'
VNET_CIDR_8 = '10.0.0.0/8'
VNET_CIDR_16 = '10.0.0.0/16'
VNET_CIDR_24 = '10.0.0.0/24'

SSH_PORT = 22
MONGODB_PORT = 27017
NTP_PORT = 123
SMTP_PORT_25 = 25
SMTP_PORT_587 = 587
HTTP_PORT = 80
HTTPS_PORT = 443
REDIS_PORT = 6379
MEMCACHED_PORT = 11211
POSTGRESQL_PORT = 5432

TCP_PROTOCOL = 6
UDP_PROTOCOL = 17
ICMP_PROTOCOL = 1
ALL_PROTOCOL = -1

#
# Azure virtual machine sizes
#

# todo add vm sizes

#
# Parameter types
#

STRING = 'String'
NUMBER = 'Number'
LIST_OF_NUMBERS = 'List<Number>'
COMMA_DELIMITED_LIST = 'CommaDelimitedList'



from compute import *

from base import ARMTemplate
from network import VirtualNetwork, AddressSpace, Subnet

template = ARMTemplate()

vnet = VirtualNetwork("myvnet",
                      addressSpace=AddressSpace(addressPrefixes=['10.0.0.0/24']),
                      subnets=[Subnet('subnet1', addressPrefix='10.0.0.0/24')])

linux_conf = LinuxConfiguration(disablePasswordAuthentication=False)

ip_conf = VirtualMachineScaleSetIPConfiguration('my_ip_config',
                                                subnet=SubResource(id=vnet.SubnetRef('subnet1')))
nic_conf = VirtualMachineScaleSetNetworkConfiguration('my_nic',
                                                      primary=True,
                                                      ipConfigurations=[ip_conf])

ss_vm_prof = VirtualMachineScaleSetVMProfile(
    osProfile=VirtualMachineScaleSetOSProfile(computerNamePrefix='testVm',
                                              adminUsername='adminuser',
                                              adminPassword='Welcome1234567+',
                                              linuxConfiguration=linux_conf),
    storageProfile=VirtualMachineScaleSetStorageProfile(imageReference=ImageReference(publisher='Canonical',
                                                                                      offer='UbuntuServer',
                                                                                      sku='16.04-LTS'),
                                                        osDisk=VirtualMachineScaleSetOSDisk(createOption='FromImage')),
    networkProfile=VirtualMachineScaleSetNetworkProfile(networkInterfaceConfigurations=[nic_conf])
)

vm_scale_set = VirtualMachineScaleSets('my_scale_set',
                                       # plan=Plan(),
                                       sku=VirtualMachineScaleSetSku(name='Standard_A1', tier='Standard', capacity=3),
                                       overprovision=True,
                                       upgradePolicy=UpgradePolicy(mode='Manual'),
                                       virtualMachineProfile=ss_vm_prof).with_depends_on(vnet)

template.add_resource([vm_scale_set, vnet])

print(template.to_json())



from base import ARMObject, ARMProperty, SubResource, SubResourceRef


class AddressSpace(ARMProperty):
    props = {
        'addressPrefixes': (list, False)
    }


class ServiceEndpointProperties(ARMProperty):
    props = {
        'service': (str, True),
        'locations': (list, True)
    }


class Subnet(ARMObject):
    props = {
        'addressPrefix': (str, True),
        'networkSecurityGroup': (SubResource, False),
        'serviceEndpoints': ((list, ServiceEndpointProperties), False)
    }


class VirtualNetwork(ARMObject):
    resource_type = 'Microsoft.Network/virtualNetworks'
    apiVersion = "2017-10-01"
    location = True

    props = {
        'addressSpace': (AddressSpace, False),
        'subnets': ((Subnet, list), False),
    }

    root_props = {
        'tags': (dict, False)
    }

    def subnet_ref(self, subnet_name):
        return "[resourceId('Microsoft.Network/virtualNetworks/subnets', '{0}', '{1}')]" \
            .format(self.title,
                    self.get_subnet(subnet_name).title)

    def get_subnet(self, subnet_name):
        return next(iter(filter(lambda x: x.title == subnet_name, self.properties['subnets'])))

    SubnetRef = subnet_ref


class ApplicationSecurityGroup(ARMObject):
    resource_type = 'Microsoft.Network/applicationSecurityGroups'
    apiVersion = "2017-10-01"
    location = True

    props = {}


class SecurityRule(ARMObject):
    props = {
        'description': (str, False),  # todo add validation on length
        'protocol': (str, True),  # todo add valiadtion on valid values: 'Tcp', 'Udp', and '*'
        'sourcePortRange': (str, False),  # todo add validation for all 'range' props
        'destinationPortRange': (str, False),
        'sourceAddressPrefix': (str, False),
    # only one of sourceAddressPrefix, sourceAddressPrefixes, sourceApplicationSecurityGroups should be non empty
        'sourceAddressPrefixes': ((list, str), False),
        'sourceApplicationSecurityGroups': ((list, SubResource), False),
        'destinationAddressPrefix': (str, False),
        'destinationAddressPrefixes': ((list, str), False),
        'destinationApplicationSecurityGroups': ((list, SubResource), False),
        'sourcePortRanges': ((list, str), False),
        'destinationPortRanges': ((list, str), False),
        'access': (str, True),  # todo add validation on values: 'Allow' and 'Deny'
        'priority': (int, False),
        'direction': (str, True)  # todo add validation on values: 'Inbound' and 'Outbound'
    }

    def validate(self):
        if 'destinationAddressPrefix' not in self.properties and \
           'destinationAddressPrefixes' not in self.properties and \
           'destinationApplicationSecurityGroups' not in self.properties:
            raise ValueError('Required security rule parameters are missing for security rule with name "{}". '
                             'Security rule must specify DestinationAddressPrefixes, DestinationAddressPrefix, or '
                             'DestinationApplicationSecurityGroups'.format(self.title))

        if 'sourceAddressPrefix' not in self.properties and \
           'sourceAddressPrefixes' not in self.properties and \
           'sourceApplicationSecurityGroups' not in self.properties:
            raise ValueError('Required security rule parameters are missing for security rule with name "{}". '
                             'Security rule must specify SourceAddressPrefixes, SourceAddressPrefix, or '
                             'SourceApplicationSecurityGroups'.format(self.title))



class PublicIPAddressDnsSettings(ARMProperty):
    props = {
        'domainNameLabel': (str, False),
        'fqdn': (str, False),
        'reverseFqdn': (str, False)
    }


class PublicIPAddress(ARMObject):
    resource_type = 'Microsoft.Network/publicIPAddresses'
    apiVersion = "2017-10-01"
    location = True

    props = {
        'publicIPAllocationMethod': (str, False),  # todo add validation on values: 'Static' and 'Dynamic'
        'publicIPAddressVersion': (str, False),  # todo add validation on values: 'IPv4' and 'IPv6'. Default is 'IPv4'
        'dnsSettings': (PublicIPAddressDnsSettings, False)
    }

    root_props = {
        'tags': (dict, False)
    }


class NetworkInterfaceIPConfiguration(ARMObject):
    props = {
        # 'applicationGatewayBackendAddressPools': () - not implemented
        # 'loadBalancerBackendAddressPools': () - not implemented
        # 'loadBalancerInboundNatRules': () - not implemented
        'privateIPAddress': (str, False),
        'privateIPAllocationMethod': (str, False),  # todo add validation on values: 'Static' and 'Dynamic'
        'privateIPAddressVersion': (str, False),  # todo add validation on values: 'IPv4' and 'IPv6'. Default is 'IPv4'
        'subnet': (SubResource, False),
        'primary': (bool, False),
        'publicIPAddress': (SubResource, False),
        'applicationSecurityGroups': ((list, SubResource), False)
    }


class NetworkInterfaceDnsSettings(ARMProperty):
    props = {
        'dnsServers': ((list, str), False),
        'appliedDnsServers': ((list, str), False),
        'internalDnsNameLabel': (str, False),
        'internalFqdn': (str, False),
        'internalDomainNameSuffix': (str, False),
    }


class NetworkSecurityGroup(ARMObject):
    resource_type = 'Microsoft.Network/networkSecurityGroups'
    apiVersion = "2017-10-01"
    location = True

    props = {
        'securityRules': ((list, SecurityRule), False)
    }

    root_props = {
        'tags': (dict, False)
    }


class NetworkInterface(ARMObject):
    resource_type = 'Microsoft.Network/networkInterfaces'
    apiVersion = "2017-10-01"
    location = True

    props = {
        'virtualMachine': (SubResource, False),
        'networkSecurityGroup': ((SubResource, NetworkSecurityGroup), False),
        'ipConfigurations': (list, True),  # type: list[NetworkInterfaceIPConfiguration]
        'dnsSettings': (NetworkInterfaceDnsSettings, False)
    }

    root_props = {
        'tags': (dict, False)
    }

    def validate(self):
        # validate ipConfigurations list item types
        if self.properties['ipConfigurations']:
            ip_confs = self.properties['ipConfigurations']
            for ip_conf in ip_confs:
                if not isinstance(ip_conf, NetworkInterfaceIPConfiguration):
                    raise ValueError("ipConfigurations in NetworkInterface must contain "
                                     "NetworkInterfaceIPConfiguration objects")


class LoadBalancerSku(ARMProperty):
    props = {'name': (str, True)}

    def validate(self):
        if self.resource['name'] not in ["Basic", "Standard"]:
            raise ValueError("Name of a load balancer SKU is not correct. Possible values are Basic or Standard")


class FrontendIPConfiguration(ARMObject):
    props = {
        'privateIPAddress': (str, False),
        'privateIPAllocationMethod': (str, False),  # Possible values are: 'Static' and 'Dynamic'
        'subnet': (SubResource, False),
        'publicIPAddress': (SubResource, False),
        'zones': ([str], False)
    }

    def to_dict(self):
        self._move_prop_to_root('zones')
        return ARMObject.to_dict(self)


class FrontendIPConfigurationRef(SubResourceRef):
    def __init__(self, load_balancer, ip_conf):
        super(FrontendIPConfigurationRef, self).__init__(LoadBalancer, 'frontendIpConfigurations', load_balancer,
                                                         ip_conf)


class ProbeRef(SubResourceRef):
    def __init__(self, load_balancer, probe):
        super(ProbeRef, self).__init__(LoadBalancer, 'probes', load_balancer, probe)


class BackendAddressPoolRef(SubResourceRef):
    def __init__(self, load_balancer, backend_pool):
        super(BackendAddressPoolRef, self).__init__(LoadBalancer, 'backendAddressPools', load_balancer, backend_pool)


class LoadBalancingRule(ARMObject):
    props = {
        'frontendIPConfiguration': (FrontendIPConfigurationRef, False),
        'backendAddressPool': (BackendAddressPoolRef, False),
        'probe': (ProbeRef, False),
        'protocol': (str, True),  # Udp, Tcp, All
        'loadDistribution': (str, False),  # Possible values are 'Default', 'SourceIP', and 'SourceIPProtocol'
        'frontendPort': (int, True),  # 0-65534, when 0 is "Any port"
        'backendPort': (int, False),  # 0-65534, when 0 is "Any port"
        'idleTimeoutInMinutes': (int, False),
    # The timeout for the TCP idle connection. The value can be set between 4 and 30 minutes. The default value is 4 minutes. This element is only used when the protocol is set to TCP.
        'enableFloatingIP': (bool, False),
        'disableOutboundSnat': (False, False),
    }


class Probe(ARMObject):
    props = {
        'protocol': (str, True),  # The protocol of the end point. Possible values are: 'Http' or 'Tcp'. If 'Tcp' is specified, a received ACK is required for the probe to be successful. If 'Http' is specified, a 200 OK response from the specifies URI is required for the probe to be successful. - Http or Tcp
        'port': (int, True),  # The port for communicating the probe. Possible values range from 1 to 65535, inclusive.
        'intervalInSeconds': (int, False),  # The interval, in seconds, for how frequently to probe the endpoint for health status. Typically, the interval is slightly less than half the allocated timeout period (in seconds) which allows two full probes before taking the instance out of rotation. The default value is 15, the minimum value is 5.
        'numberOfProbes': (int, False),  # The number of probes where if no response, will result in stopping further traffic from being delivered to the endpoint. This values allows endpoints to be taken out of rotation faster or slower than the typical times used in Azure.
        'requestPath': (str, False),  # The URI used for requesting health status from the VM. Path is required if a protocol is set to http. Otherwise, it is not allowed. There is no default value.
    }


class BackendAddressPool(ARMProperty):
    props = {'name': (str, True)}


class LoadBalancer(ARMObject):
    resource_type = 'Microsoft.Network/loadBalancers'
    apiVersion = "2017-10-01"
    location = True

    props = {
        'frontendIPConfigurations': ([FrontendIPConfiguration], False),
        'backendAddressPools': ([BackendAddressPool], False),
        'loadBalancingRules': ([LoadBalancingRule], False),
        'probes': ([Probe], False),
        # 'inboundNatRules': ([InboundNatRule], False),
        # 'inboundNatPools': ([InboundNatPool], False),
        # 'outboundNatRules': ([OutboundNatRule], False),
        # 'resources': [inboundNatRules]
    }

    root_props = {
        'sku': (LoadBalancerSku, True),
        'tags': (dict, False)
    }



import uuid
from compute import *
from base import ARMTemplate, ARMParameter
from network import *
from storage import StorageAccount, StorageAccountSku

template = ARMTemplate()

vm_password_param = ARMParameter('vmPassword',
                                 type='secureString',
                                 description='The password for the VM access. User is "adminuser"')
template.add_parameter(vm_password_param)

vnet = VirtualNetwork("myvnet",
                      addressSpace=AddressSpace(addressPrefixes=['10.0.0.0/24']),
                      subnets=[Subnet('subnet1', addressPrefix='10.0.0.0/24')],
                      tags={'Key1': 'tag-key1', 'Key2': 'tag-key2', 'Key3': 'tag-key3'})

nsg = NetworkSecurityGroup("testvmNsg",
                           securityRules=[
                               SecurityRule('ssh',
                                            description='public ssh access',
                                            protocol='Tcp',
                                            destinationPortRange="22",
                                            destinationAddressPrefix="*",
                                            sourceAddressPrefix='Internet',
                                            sourcePortRange='*',
                                            access='Allow',
                                            direction='Inbound',
                                            priority=200
                                            )])

publicIp = PublicIPAddress('testvm_nic1_pubip', publicIPAllocationMethod='dynamic')

networkInterface = NetworkInterface('tesmvm_nic1',
                                    ipConfigurations=
                                        [NetworkInterfaceIPConfiguration('tesmvm_nic1_ip_config',
                                                                     privateIPAllocationMethod='Dynamic',
                                                                     subnet=SubResource(id=vnet.SubnetRef('subnet1')),
                                                                     publicIPAddress=SubResource(id=publicIp.Ref()))],
                                    networkSecurityGroup=SubResource(id=nsg.Ref()))
networkInterface.with_depends_on([publicIp, vnet])


vm = VirtualMachine('testvm',
                    hardwareProfile=HardwareProfile(vmSize='Basic_A0'),
                    storageProfile=StorageProfile(imageReference=ImageReference(publisher='Canonical',
                                                                                offer='UbuntuServer',
                                                                                sku='16.04-LTS'),
                                                  osDisk=OsDisk(createOption='FromImage',
                                                                diskSizeGB=50,
                                                                managedDisk=ManagedDiskParameters(storageAccountType='Standard_LRS'))),
                    osProfile=OSProfile(computerName='alextesm',
                                        adminUsername='adminuser',
                                        adminPassword=vm_password_param.Ref(),
                                        linuxConfiguration=LinuxConfiguration(disablePasswordAuthentication=False)),
                    networkProfile=NetworkProfile(networkInterfaces=
                                                  [NetworkInterfaceReference(id=networkInterface.Ref())]))
vm.with_depends_on(networkInterface)

# create VM extension
ext_script = VirtualMachineExtension.create_linux_custom_script(
    vm=vm,
    command_to_execute="sh some_script.sh",
    file_uris="https://somescript.sh",
    protected_settings=True)


storage = StorageAccount(str(uuid.uuid4())[-12:],
                         sku=StorageAccountSku(name='Standard_LRS'),
                         kind='StorageV2',
                         tags={'tag1': 'bla1'})

template.add_resource([vnet, nsg, publicIp, networkInterface, vm, storage, ext_script])

print(template.to_json())



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name='ionosphere',
    version="0.0.18",
    description="Azure Resource Manager Template creation library",
    long_description=readme(),
    author="Alex Azarh",
    author_email="alex.az@quali.com",
    url="https://github.com/qualinext/ionosphere",
    license="New BSD license",
    packages=['ionosphere', 'ionosphere.helpers'],
    scripts=['scripts/cfn', 'scripts/cfn2py'],
    install_requires=["cfn_flip>=0.2.5"],
    test_suite="tests",
    tests_require=[],
    extras_require={'policy': ['awacs']},
    use_2to3=True,
)
